#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import http.client
import uuid

from oslo_serialization import jsonutils

from keystone.common.policies import role_assignment as rp
from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class _AssignmentTestUtilities:
    """Useful utilities for setting up test assignments and assertions."""

    def _setup_test_role_assignments(self):
        # Utility to create assignments and return important data for
        # assertions.

        # Since the role doesn't really matter too much, we can just re-use an
        # existing role instead of creating a new one.
        role_id = self.bootstrapper.reader_role_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        # create a user+project role assignment.
        PROVIDERS.assignment_api.create_grant(
            role_id, user_id=user['id'], project_id=project['id']
        )

        # create a user+domain role assignment.
        PROVIDERS.assignment_api.create_grant(
            role_id, user_id=user['id'], domain_id=domain['id']
        )

        # create a user+system role assignment.
        PROVIDERS.assignment_api.create_system_grant_for_user(
            user['id'], role_id
        )

        # create a group+project role assignment.
        PROVIDERS.assignment_api.create_grant(
            role_id, group_id=group['id'], project_id=project['id']
        )

        # create a group+domain role assignment.
        PROVIDERS.assignment_api.create_grant(
            role_id, group_id=group['id'], domain_id=domain['id']
        )

        # create a group+system role assignment.
        PROVIDERS.assignment_api.create_system_grant_for_group(
            group['id'], role_id
        )

        return {
            'user_id': user['id'],
            'group_id': group['id'],
            'domain_id': domain['id'],
            'project_id': project['id'],
            'role_id': role_id,
        }

    def _extract_role_assignments_from_response_body(self, r):
        # Condense the role assignment details into a set of key things we can
        # use in assertions.
        assignments = []
        for assignment in r.json['role_assignments']:
            a = {}
            if 'project' in assignment['scope']:
                a['project_id'] = assignment['scope']['project']['id']
            elif 'domain' in assignment['scope']:
                a['domain_id'] = assignment['scope']['domain']['id']
            elif 'system' in assignment['scope']:
                a['system'] = 'all'

            if 'user' in assignment:
                a['user_id'] = assignment['user']['id']
            elif 'group' in assignment:
                a['group_id'] = assignment['group']['id']

            a['role_id'] = assignment['role']['id']

            assignments.append(a)
        return assignments


class _SystemUserTests:
    """Common functionality for system users regardless of default role."""

    def test_user_can_list_all_role_assignments_in_the_deployment(self):
        assignments = self._setup_test_role_assignments()

        # this assignment is created by keystone-manage bootstrap
        self.expected.append(
            {
                'user_id': self.bootstrapper.admin_user_id,
                'project_id': self.bootstrapper.project_id,
                'role_id': self.bootstrapper.admin_role_id,
            }
        )

        # this assignment is created by keystone-manage bootstrap
        self.expected.append(
            {
                'user_id': self.bootstrapper.admin_user_id,
                'system': 'all',
                'role_id': self.bootstrapper.admin_role_id,
            }
        )
        self.expected.append(
            {
                'user_id': assignments['user_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'user_id': assignments['user_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'user_id': assignments['user_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': assignments['group_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': assignments['group_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': assignments['group_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            }
        )

        with self.test_client() as c:
            r = c.get('/v3/role_assignments', headers=self.headers)
            self.assertEqual(
                len(self.expected), len(r.json['role_assignments'])
            )
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, self.expected)

    def test_user_can_list_all_role_names_assignments_in_the_deployment(self):
        assignments = self._setup_test_role_assignments()

        # this assignment is created by keystone-manage bootstrap
        self.expected.append(
            {
                'user_id': self.bootstrapper.admin_user_id,
                'project_id': self.bootstrapper.project_id,
                'role_id': self.bootstrapper.admin_role_id,
            }
        )

        # this assignment is created by keystone-manage bootstrap
        self.expected.append(
            {
                'user_id': self.bootstrapper.admin_user_id,
                'system': 'all',
                'role_id': self.bootstrapper.admin_role_id,
            }
        )
        self.expected.append(
            {
                'user_id': assignments['user_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'user_id': assignments['user_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'user_id': assignments['user_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': assignments['group_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': assignments['group_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': assignments['group_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            }
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/role_assignments?include_names=True', headers=self.headers
            )
            self.assertEqual(
                len(self.expected), len(r.json['role_assignments'])
            )
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, self.expected)

    def test_user_can_filter_role_assignments_by_project(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            {
                'user_id': assignments['user_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            },
            {
                'group_id': assignments['group_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            },
        ]
        project_id = assignments['project_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?scope.project.id={project_id}',
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_domain(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            {
                'user_id': assignments['user_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            },
            {
                'group_id': assignments['group_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            },
        ]
        domain_id = assignments['domain_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?scope.domain.id={domain_id}',
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_system(self):
        assignments = self._setup_test_role_assignments()

        # this assignment is created by keystone-manage bootstrap
        self.expected.append(
            {
                'user_id': self.bootstrapper.admin_user_id,
                'system': 'all',
                'role_id': self.bootstrapper.admin_role_id,
            }
        )
        self.expected.append(
            {
                'user_id': assignments['user_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': assignments['group_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            }
        )

        with self.test_client() as c:
            r = c.get(
                '/v3/role_assignments?scope.system=all', headers=self.headers
            )
            self.assertEqual(
                len(self.expected), len(r.json['role_assignments'])
            )
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, self.expected)

    def test_user_can_filter_role_assignments_by_user(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            # assignment of the user running the test case
            {
                'user_id': assignments['user_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            },
            {
                'user_id': assignments['user_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            },
            {
                'user_id': assignments['user_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            },
        ]
        user_id = assignments['user_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?user.id={user_id}', headers=self.headers
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_group(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            {
                'group_id': assignments['group_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            },
            {
                'group_id': assignments['group_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            },
            {
                'group_id': assignments['group_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            },
        ]
        group_id = assignments['group_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?group.id={group_id}',
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_role(self):
        assignments = self._setup_test_role_assignments()
        self.expected = [
            ra
            for ra in self.expected
            if ra['role_id'] == assignments['role_id']
        ]
        self.expected.append(
            {
                'user_id': assignments['user_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'user_id': assignments['user_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'user_id': assignments['user_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': assignments['group_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': assignments['group_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': assignments['group_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            }
        )

        role_id = assignments['role_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?role.id={role_id}&include_names=True',
                headers=self.headers,
            )
            self.assertEqual(
                len(self.expected), len(r.json['role_assignments'])
            )
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, self.expected)

    def test_user_can_filter_role_assignments_by_project_and_role(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            {
                'user_id': assignments['user_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            },
            {
                'group_id': assignments['group_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            },
        ]

        with self.test_client() as c:
            qs = (assignments['project_id'], assignments['role_id'])
            r = c.get(
                '/v3/role_assignments?scope.project.id={}&role.id={}'.format(
                    *qs
                ),
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_domain_and_role(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            {
                'user_id': assignments['user_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            },
            {
                'group_id': assignments['group_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            },
        ]
        qs = (assignments['domain_id'], assignments['role_id'])

        with self.test_client() as c:
            r = c.get(
                '/v3/role_assignments?scope.domain.id={}&role.id={}'.format(
                    *qs
                ),
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_system_and_role(self):
        assignments = self._setup_test_role_assignments()
        self.expected = [
            ra
            for ra in self.expected
            if ra['role_id'] == assignments['role_id']
        ]
        self.expected.append(
            {
                'user_id': assignments['user_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': assignments['group_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            }
        )
        role_id = assignments['role_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?scope.system=all&role.id={role_id}',
                headers=self.headers,
            )
            self.assertEqual(
                len(self.expected), len(r.json['role_assignments'])
            )
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, self.expected)

    def test_user_can_filter_role_assignments_by_user_and_role(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            {
                'user_id': assignments['user_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            },
            {
                'user_id': assignments['user_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            },
            {
                'user_id': assignments['user_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            },
        ]
        qs = (assignments['user_id'], assignments['role_id'])

        with self.test_client() as c:
            r = c.get(
                '/v3/role_assignments?user.id={}&role.id={}'.format(*qs),
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_group_and_role(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            {
                'group_id': assignments['group_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            },
            {
                'group_id': assignments['group_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            },
            {
                'group_id': assignments['group_id'],
                'system': 'all',
                'role_id': assignments['role_id'],
            },
        ]

        with self.test_client() as c:
            qs = (assignments['group_id'], assignments['role_id'])
            r = c.get(
                '/v3/role_assignments?group.id={}&role.id={}'.format(*qs),
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_project_and_user(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            {
                'user_id': assignments['user_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            }
        ]
        qs = (assignments['project_id'], assignments['user_id'])

        with self.test_client() as c:
            r = c.get(
                '/v3/role_assignments?scope.project.id={}&user.id={}'.format(
                    *qs
                ),
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_project_and_group(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            {
                'group_id': assignments['group_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            }
        ]
        qs = (assignments['project_id'], assignments['group_id'])

        with self.test_client() as c:
            r = c.get(
                '/v3/role_assignments?scope.project.id={}&group.id={}'.format(
                    *qs
                ),
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_domain_and_user(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            {
                'user_id': assignments['user_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            }
        ]
        qs = (assignments['domain_id'], assignments['user_id'])

        with self.test_client() as c:
            r = c.get(
                '/v3/role_assignments?scope.domain.id={}&user.id={}'.format(
                    *qs
                ),
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_domain_and_group(self):
        assignments = self._setup_test_role_assignments()
        expected = [
            {
                'group_id': assignments['group_id'],
                'domain_id': assignments['domain_id'],
                'role_id': assignments['role_id'],
            }
        ]
        qs = (assignments['domain_id'], assignments['group_id'])

        with self.test_client() as c:
            r = c.get(
                '/v3/role_assignments?scope.domain.id={}&group.id={}'.format(
                    *qs
                ),
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_list_assignments_for_subtree(self):
        assignments = self._setup_test_role_assignments()
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(
                domain_id=CONF.identity.default_domain_id,
                parent_id=assignments['project_id'],
            ),
        )
        PROVIDERS.assignment_api.create_grant(
            assignments['role_id'],
            user_id=user['id'],
            project_id=project['id'],
        )
        expected = [
            {
                'user_id': assignments['user_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            },
            {
                'group_id': assignments['group_id'],
                'project_id': assignments['project_id'],
                'role_id': assignments['role_id'],
            },
            {
                'user_id': user['id'],
                'project_id': project['id'],
                'role_id': assignments['role_id'],
            },
        ]
        with self.test_client() as c:
            r = c.get(
                (
                    '/v3/role_assignments?scope.project.id={}&include_subtree'.format(
                        assignments['project_id']
                    )
                ),
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)


class _DomainUserTests:
    """Common functionality for domain users."""

    def _setup_test_role_assignments_for_domain(self):
        # Populate role assignment within `self.domain_id` so that we can
        # assert users can view assignments within the domain they have
        # authorization on
        role_id = self.bootstrapper.reader_role_id

        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=CONF.identity.default_domain_id)
        )

        group = PROVIDERS.identity_api.create_group(
            unit.new_group_ref(domain_id=CONF.identity.default_domain_id)
        )

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        # create a user+project role assignment.
        PROVIDERS.assignment_api.create_grant(
            role_id, user_id=user['id'], project_id=project['id']
        )

        # create a user+domain role assignment.
        PROVIDERS.assignment_api.create_grant(
            role_id, user_id=user['id'], domain_id=self.domain_id
        )

        # create a group+project role assignment.
        PROVIDERS.assignment_api.create_grant(
            role_id, group_id=group['id'], project_id=project['id']
        )

        # create a group+domain role assignment.
        PROVIDERS.assignment_api.create_grant(
            role_id, group_id=group['id'], domain_id=self.domain_id
        )

        return {
            'user_id': user['id'],
            'group_id': group['id'],
            'project_id': project['id'],
            'role_id': role_id,
        }

    def test_user_can_list_all_assignments_in_their_domain(self):
        self._setup_test_role_assignments()
        domain_assignments = self._setup_test_role_assignments_for_domain()

        self.expected.append(
            {
                'user_id': domain_assignments['user_id'],
                'domain_id': self.domain_id,
                'role_id': domain_assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'user_id': domain_assignments['user_id'],
                'project_id': domain_assignments['project_id'],
                'role_id': domain_assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': domain_assignments['group_id'],
                'domain_id': self.domain_id,
                'role_id': domain_assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': domain_assignments['group_id'],
                'project_id': domain_assignments['project_id'],
                'role_id': domain_assignments['role_id'],
            }
        )

        with self.test_client() as c:
            r = c.get('/v3/role_assignments', headers=self.headers)
            self.assertEqual(
                len(self.expected), len(r.json['role_assignments'])
            )
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, self.expected)

    def test_user_can_filter_role_assignments_by_project_in_domain(self):
        self._setup_test_role_assignments()
        domain_assignments = self._setup_test_role_assignments_for_domain()

        expected = [
            {
                'user_id': domain_assignments['user_id'],
                'project_id': domain_assignments['project_id'],
                'role_id': domain_assignments['role_id'],
            },
            {
                'group_id': domain_assignments['group_id'],
                'project_id': domain_assignments['project_id'],
                'role_id': domain_assignments['role_id'],
            },
        ]

        project_id = domain_assignments['project_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?scope.project.id={project_id}',
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_domain(self):
        # This shouldn't really provide any more value than just calling GET
        # /v3/role_assignments with a domain-scoped token, but we test it
        # anyway.
        self._setup_test_role_assignments()
        domain_assignments = self._setup_test_role_assignments_for_domain()

        self.expected.append(
            {
                'user_id': domain_assignments['user_id'],
                'domain_id': self.domain_id,
                'role_id': domain_assignments['role_id'],
            }
        )
        self.expected.append(
            {
                'group_id': domain_assignments['group_id'],
                'domain_id': self.domain_id,
                'role_id': domain_assignments['role_id'],
            }
        )

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?scope.domain.id={self.domain_id}',
                headers=self.headers,
            )
            self.assertEqual(
                len(self.expected), len(r.json['role_assignments'])
            )
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, self.expected)

    def test_user_can_filter_role_assignments_by_user_of_domain(self):
        self._setup_test_role_assignments()
        domain_assignments = self._setup_test_role_assignments_for_domain()

        expected = [
            {
                'user_id': domain_assignments['user_id'],
                'domain_id': self.domain_id,
                'role_id': domain_assignments['role_id'],
            },
            {
                'user_id': domain_assignments['user_id'],
                'project_id': domain_assignments['project_id'],
                'role_id': domain_assignments['role_id'],
            },
        ]

        user_id = domain_assignments['user_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?user.id={user_id}', headers=self.headers
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_can_filter_role_assignments_by_group_of_domain(self):
        self._setup_test_role_assignments()
        domain_assignments = self._setup_test_role_assignments_for_domain()

        expected = [
            {
                'group_id': domain_assignments['group_id'],
                'domain_id': self.domain_id,
                'role_id': domain_assignments['role_id'],
            },
            {
                'group_id': domain_assignments['group_id'],
                'project_id': domain_assignments['project_id'],
                'role_id': domain_assignments['role_id'],
            },
        ]

        group_id = domain_assignments['group_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?group.id={group_id}',
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_cannot_filter_role_assignments_by_system(self):
        self._setup_test_role_assignments()
        self._setup_test_role_assignments_for_domain()

        with self.test_client() as c:
            r = c.get(
                '/v3/role_assignments?scope.system=all', headers=self.headers
            )
            self.assertEqual(0, len(r.json['role_assignments']))

    def test_user_cannot_filter_role_assignments_by_other_domain(self):
        assignments = self._setup_test_role_assignments()
        domain = assignments['domain_id']
        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?scope.domain.id={domain}',
                headers=self.headers,
            )
            self.assertEqual([], r.json['role_assignments'])

    def test_user_cannot_filter_role_assignments_by_other_domain_project(self):
        assignments = self._setup_test_role_assignments()
        self._setup_test_role_assignments_for_domain()

        # This project is in an entirely separate domain that this user doesn't
        # have authorization to access, so they should only see an empty list
        project_id = assignments['project_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?scope.project.id={project_id}',
                headers=self.headers,
            )
            self.assertEqual(0, len(r.json['role_assignments']))

    def test_user_cannot_filter_role_assignments_by_other_domain_user(self):
        assignments = self._setup_test_role_assignments()
        self._setup_test_role_assignments_for_domain()

        # This user doesn't have any role assignments on self.domain_id, so the
        # domain user of self.domain_id should only see an empty list of role
        # assignments.
        user_id = assignments['user_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?user.id={user_id}', headers=self.headers
            )
            self.assertEqual(0, len(r.json['role_assignments']))

    def test_user_cannot_filter_role_assignments_by_other_domain_group(self):
        assignments = self._setup_test_role_assignments()
        self._setup_test_role_assignments_for_domain()

        # This group doesn't have any role assignments on self.domain_id, so
        # the domain user of self.domain_id should only see an empty list of
        # role assignments.
        group_id = assignments['group_id']

        with self.test_client() as c:
            r = c.get(
                f'/v3/role_assignments?group.id={group_id}',
                headers=self.headers,
            )
            self.assertEqual(0, len(r.json['role_assignments']))

    def test_user_can_list_assignments_for_subtree_in_their_domain(self):
        assignments = self._setup_test_role_assignments()
        domain_assignments = self._setup_test_role_assignments_for_domain()
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(
                domain_id=self.domain_id,
                parent_id=domain_assignments['project_id'],
            ),
        )
        PROVIDERS.assignment_api.create_grant(
            assignments['role_id'],
            user_id=user['id'],
            project_id=project['id'],
        )
        expected = [
            {
                'user_id': domain_assignments['user_id'],
                'project_id': domain_assignments['project_id'],
                'role_id': assignments['role_id'],
            },
            {
                'group_id': domain_assignments['group_id'],
                'project_id': domain_assignments['project_id'],
                'role_id': assignments['role_id'],
            },
            {
                'user_id': user['id'],
                'project_id': project['id'],
                'role_id': assignments['role_id'],
            },
        ]
        with self.test_client() as c:
            r = c.get(
                (
                    '/v3/role_assignments?scope.project.id={}&include_subtree'.format(
                        domain_assignments['project_id']
                    )
                ),
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_cannot_list_assignments_for_subtree_in_other_domain(self):
        assignments = self._setup_test_role_assignments()
        with self.test_client() as c:
            c.get(
                (
                    '/v3/role_assignments?scope.project.id={}&include_subtree'.format(
                        assignments['project_id']
                    )
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class _ProjectUserTests:
    def test_user_cannot_list_all_assignments_in_their_project(self):
        with self.test_client() as c:
            c.get(
                '/v3/role_assignments',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_filter_role_assignments_by_user_of_project(self):
        assignments = self._setup_test_role_assignments()
        user_id = assignments['user_id']

        with self.test_client() as c:
            c.get(
                f'/v3/role_assignments?user.id={user_id}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_filter_role_assignments_by_group_of_project(self):
        assignments = self._setup_test_role_assignments()
        group_id = assignments['group_id']

        with self.test_client() as c:
            c.get(
                f'/v3/role_assignments?group.id={group_id}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_filter_role_assignments_by_system(self):
        with self.test_client() as c:
            c.get(
                '/v3/role_assignments?scope.system=all',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_filter_role_assignments_by_domain(self):
        with self.test_client() as c:
            c.get(
                f'/v3/role_assignments?scope.domain.id={self.domain_id}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_filter_role_assignments_by_other_project(self):
        project1 = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        with self.test_client() as c:
            c.get(
                f'/v3/role_assignments?scope.project.id={project1}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_filter_role_assignments_by_other_project_user(self):
        assignments = self._setup_test_role_assignments()

        # This user doesn't have any role assignments on self.project_id, so
        # the project user of self.project_id should only see an empty list of
        # role assignments.
        user_id = assignments['user_id']

        with self.test_client() as c:
            c.get(
                f'/v3/role_assignments?user.id={user_id}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_filter_role_assignments_by_other_project_group(self):
        assignments = self._setup_test_role_assignments()

        # This group doesn't have any role assignments on self.project_id, so
        # the project user of self.project_id should only see an empty list of
        # role assignments.
        group_id = assignments['group_id']

        with self.test_client() as c:
            c.get(
                f'/v3/role_assignments?group.id={group_id}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class _ProjectReaderMemberTests:
    def test_user_cannot_list_assignments_for_subtree(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(
                domain_id=self.domain_id, parent_id=self.project_id
            ),
        )
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )
        with self.test_client() as c:
            c.get(
                (
                    f'/v3/role_assignments?scope.project.id={self.project_id}&include_subtree'
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class SystemReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _AssignmentTestUtilities,
    _SystemUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_reader)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )
        self.expected = [
            # assignment of the user running the test case
            {
                'user_id': self.user_id,
                'system': 'all',
                'role_id': self.bootstrapper.reader_role_id,
            }
        ]

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_reader['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _AssignmentTestUtilities,
    _SystemUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_member)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )
        self.expected = [
            # assignment of the user running the test case
            {
                'user_id': self.user_id,
                'system': 'all',
                'role_id': self.bootstrapper.member_role_id,
            }
        ]

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_member['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _AssignmentTestUtilities,
    _SystemUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        self.user_id = self.bootstrapper.admin_user_id
        self.expected = []

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class DomainReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _AssignmentTestUtilities,
    _DomainUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_reader = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_reader)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )
        self.expected = [
            # assignment of the user running the test case
            {
                'user_id': self.user_id,
                'domain_id': self.domain_id,
                'role_id': self.bootstrapper.reader_role_id,
            }
        ]

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_reader['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class DomainMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _AssignmentTestUtilities,
    _DomainUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_user = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_user)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )
        self.expected = [
            # assignment of the user running the test case
            {
                'user_id': self.user_id,
                'domain_id': self.domain_id,
                'role_id': self.bootstrapper.member_role_id,
            }
        ]

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_user['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class DomainAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _AssignmentTestUtilities,
    _DomainUserTests,
):
    def _override_policy(self):
        # TODO(lbragstad): Remove this once the deprecated policies in
        # keystone.common.policies.role_assignment have been removed. This is
        # only here to make sure we test the new policies instead of the
        # deprecated ones. Oslo.policy will OR deprecated policies with new
        # policies to maintain compatibility and give operators a chance to
        # update permissions or update policies without breaking users. This
        # will cause these specific tests to fail since we're trying to correct
        # this broken behavior with better scope checking.
        with open(self.policy_file_name, 'w') as f:
            overridden_policies = {
                'identity:list_role_assignments': (
                    rp.SYSTEM_READER_OR_DOMAIN_READER
                ),
                'identity:list_role_assignments_for_tree': (
                    rp.SYSTEM_READER_OR_PROJECT_DOMAIN_READER_OR_PROJECT_ADMIN
                ),
            }
            f.write(jsonutils.dumps(overridden_policies))

    def setUp(self):
        super().setUp()
        self.loadapp()
        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_admin = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.admin_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )
        self.expected = [
            # assignment of the user running the test case
            {
                'user_id': self.user_id,
                'domain_id': self.domain_id,
                'role_id': self.bootstrapper.admin_role_id,
            }
        ]

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _AssignmentTestUtilities,
    _ProjectUserTests,
    _ProjectReaderMemberTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']

        project = unit.new_project_ref(domain_id=self.domain_id)
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        self.project_id = project['id']

        project_reader = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(project_reader)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=self.user_id,
            project_id=self.project_id,
        )

        self.expected = [
            # assignment of the user running the test case
            {
                'user_id': self.user_id,
                'project_id': self.project_id,
                'role_id': self.bootstrapper.reader_role_id,
            }
        ]

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_reader['password'],
            project_id=self.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _AssignmentTestUtilities,
    _ProjectUserTests,
    _ProjectReaderMemberTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']

        project = unit.new_project_ref(domain_id=self.domain_id)
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        self.project_id = project['id']

        project_member = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(project_member)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            project_id=self.project_id,
        )

        self.expected = [
            # assignment of the user running the test case
            {
                'user_id': self.user_id,
                'project_id': self.project_id,
                'role_id': self.bootstrapper.member_role_id,
            }
        ]

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=project_member['password'],
            project_id=self.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _AssignmentTestUtilities,
    _ProjectUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        self._override_policy()
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']

        self.user_id = self.bootstrapper.admin_user_id

        project = unit.new_project_ref(domain_id=self.domain_id)
        project = PROVIDERS.resource_api.create_project(project['id'], project)
        self.project_id = project['id']

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.admin_role_id,
            user_id=self.user_id,
            project_id=self.project_id,
        )

        self.expected = [
            # assignment of the user running the test case
            {
                'user_id': self.user_id,
                'project_id': self.project_id,
                'role_id': self.bootstrapper.admin_role_id,
            }
        ]

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            project_id=self.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def _override_policy(self):
        # TODO(lbragstad): Remove this once the deprecated policies in
        # keystone.common.policies.role_assignment have been removed. This is
        # only here to make sure we test the new policies instead of the
        # deprecated ones. Oslo.policy will OR deprecated policies with new
        # policies to maintain compatibility and give operators a chance to
        # update permissions or update policies without breaking users. This
        # will cause these specific tests to fail since we're trying to correct
        # this broken behavior with better scope checking.
        with open(self.policy_file_name, 'w') as f:
            overridden_policies = {
                'identity:list_role_assignments': (
                    rp.SYSTEM_READER_OR_DOMAIN_READER
                ),
                'identity:list_role_assignments_for_tree': (
                    rp.SYSTEM_READER_OR_PROJECT_DOMAIN_READER_OR_PROJECT_ADMIN
                ),
            }
            f.write(jsonutils.dumps(overridden_policies))

    def test_user_can_list_assignments_for_subtree_on_own_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(
                domain_id=self.domain_id, parent_id=self.project_id
            ),
        )
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )
        expected = copy.copy(self.expected)
        expected.append(
            {
                'project_id': project['id'],
                'user_id': user['id'],
                'role_id': self.bootstrapper.reader_role_id,
            }
        )
        with self.test_client() as c:
            r = c.get(
                (
                    f'/v3/role_assignments?scope.project.id={self.project_id}&include_subtree'
                ),
                headers=self.headers,
            )
            self.assertEqual(len(expected), len(r.json['role_assignments']))
            actual = self._extract_role_assignments_from_response_body(r)
            for assignment in actual:
                self.assertIn(assignment, expected)

    def test_user_cannot_list_assignments_for_subtree_on_other_project(self):
        user = PROVIDERS.identity_api.create_user(
            unit.new_user_ref(domain_id=self.domain_id)
        )
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=user['id'],
            project_id=project['id'],
        )
        with self.test_client() as c:
            c.get(
                (
                    '/v3/role_assignments?scope.project.id={}&include_subtree'.format(
                        project['id']
                    )
                ),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )



# Copyright 2013 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Main entry point into the Credential service."""

import json

from keystone.common import cache
from keystone.common import driver_hints
from keystone.common import manager
from keystone.common import provider_api
import keystone.conf
from keystone import exception
from keystone import notifications

CONF = keystone.conf.CONF
MEMOIZE = cache.get_memoization_decorator(group='credential')
PROVIDERS = provider_api.ProviderAPIs


class Manager(manager.Manager):
    """Default pivot point for the Credential backend.

    See :mod:`keystone.common.manager.Manager` for more details on how this
    dynamically calls the backend.

    """

    driver_namespace = 'keystone.credential'
    _provides_api = 'credential_api'

    _CRED = 'credential'

    def __init__(self):
        super().__init__(CONF.credential.driver)

    def _decrypt_credential(self, credential):
        """Return a decrypted credential reference."""
        if credential['type'] == 'ec2':
            decrypted_blob = json.loads(
                PROVIDERS.credential_provider_api.decrypt(
                    credential['encrypted_blob']
                )
            )
        else:
            decrypted_blob = PROVIDERS.credential_provider_api.decrypt(
                credential['encrypted_blob']
            )
        credential['blob'] = decrypted_blob
        credential.pop('key_hash', None)
        credential.pop('encrypted_blob', None)
        return credential

    def _encrypt_credential(self, credential):
        """Return an encrypted credential reference."""
        credential_copy = credential.copy()
        if credential.get('type', None) == 'ec2':
            # NOTE(lbragstad): When dealing with ec2 credentials, it's possible
            # for the `blob` to be a dictionary. Let's make sure we are
            # encrypting a string otherwise encryption will fail.
            encrypted_blob, key_hash = (
                PROVIDERS.credential_provider_api.encrypt(
                    json.dumps(credential['blob'])
                )
            )
        else:
            encrypted_blob, key_hash = (
                PROVIDERS.credential_provider_api.encrypt(credential['blob'])
            )
        credential_copy['encrypted_blob'] = encrypted_blob
        credential_copy['key_hash'] = key_hash
        credential_copy.pop('blob', None)
        return credential_copy

    def _assert_limit_not_exceeded(self, user_id):
        user_limit = CONF.credential.user_limit
        if user_limit >= 0:
            cred_count = len(self.list_credentials_for_user(user_id))
            if cred_count >= user_limit:
                raise exception.CredentialLimitExceeded(limit=user_limit)

    @manager.response_truncated
    def list_credentials(self, hints=None):
        credentials = self.driver.list_credentials(
            hints or driver_hints.Hints()
        )
        for credential in credentials:
            credential = self._decrypt_credential(credential)
        return credentials

    def list_credentials_for_user(self, user_id, type=None):
        credentials = self._list_credentials_for_user(user_id, type)
        for credential in credentials:
            credential = self._decrypt_credential(credential)
        return credentials

    @MEMOIZE
    def _list_credentials_for_user(self, user_id, type):
        """List credentials for a specific user."""
        return self.driver.list_credentials_for_user(user_id, type)

    def get_credential(self, credential_id):
        """Return a credential reference."""
        credential = self._get_credential(credential_id)
        return self._decrypt_credential(credential)

    @MEMOIZE
    def _get_credential(self, credential_id):
        return self.driver.get_credential(credential_id)

    def create_credential(self, credential_id, credential, initiator=None):
        """Create a credential."""
        credential_copy = self._encrypt_credential(credential)
        user_id = credential_copy['user_id']
        self._assert_limit_not_exceeded(user_id)
        ref = self.driver.create_credential(credential_id, credential_copy)
        if MEMOIZE.should_cache(ref):
            self._get_credential.set(ref, credential_copy, credential_id)
            self._list_credentials_for_user.invalidate(
                self, ref['user_id'], ref['type']
            )
            self._list_credentials_for_user.invalidate(
                self, ref['user_id'], None
            )
        ref.pop('key_hash', None)
        ref.pop('encrypted_blob', None)
        ref['blob'] = credential['blob']
        notifications.Audit.created(self._CRED, credential_id, initiator)
        return ref

    def _validate_credential_update(self, credential_id, credential):
        # ec2 credentials require a "project_id" to be functional. Before we
        # update, check the case where a non-ec2 credential changes its type
        # to be "ec2", but has no associated "project_id", either in the
        # request or already set in the database
        if credential.get('type', '').lower() == 'ec2' and not credential.get(
            'project_id'
        ):
            existing_cred = self.get_credential(credential_id)
            if not existing_cred['project_id']:
                raise exception.ValidationError(
                    attribute='project_id', target='credential'
                )

    def update_credential(self, credential_id, credential):
        """Update an existing credential."""
        self._validate_credential_update(credential_id, credential)
        if 'blob' in credential:
            credential_copy = self._encrypt_credential(credential)
        else:
            credential_copy = credential.copy()
            existing_credential = self.get_credential(credential_id)
            existing_blob = existing_credential['blob']
        ref = self.driver.update_credential(credential_id, credential_copy)
        if MEMOIZE.should_cache(ref):
            self._get_credential.set(ref, self, credential_id)
            self._list_credentials_for_user.invalidate(
                self, ref['user_id'], ref['type']
            )
            self._list_credentials_for_user.invalidate(
                self, ref['user_id'], None
            )
        ref.pop('key_hash', None)
        ref.pop('encrypted_blob', None)
        # If the update request contains a `blob` attribute - we should return
        # that in the update response. If not, then we should return the
        # existing `blob` attribute since it wasn't updated.
        if credential.get('blob'):
            ref['blob'] = credential['blob']
        else:
            ref['blob'] = existing_blob
        return ref

    def delete_credential(self, credential_id, initiator=None):
        """Delete a credential."""
        cred = self.get_credential(credential_id)
        self.driver.delete_credential(credential_id)
        self._get_credential.invalidate(self, credential_id)
        self._list_credentials_for_user.invalidate(
            self, cred['user_id'], cred['type']
        )
        self._list_credentials_for_user.invalidate(self, cred['user_id'], None)
        notifications.Audit.deleted(self._CRED, credential_id, initiator)

    def delete_credentials_for_project(self, project_id):
        """Delete all credentials for a project."""
        hints = driver_hints.Hints()
        hints.add_filter('project_id', project_id)
        creds = self.driver.list_credentials(hints)

        self.driver.delete_credentials_for_project(project_id)
        for cred in creds:
            self._get_credential.invalidate(self, cred['id'])
            self._list_credentials_for_user.invalidate(
                self, cred['user_id'], cred['type']
            )
            self._list_credentials_for_user.invalidate(
                self, cred['user_id'], None
            )

    def delete_credentials_for_user(self, user_id):
        """Delete all credentials for a user."""
        creds = self.driver.list_credentials_for_user(user_id)
        self.driver.delete_credentials_for_user(user_id)
        for cred in creds:
            self._get_credential.invalidate(self, cred['id'])
            self._list_credentials_for_user.invalidate(
                self, user_id, cred['type']
            )
            self._list_credentials_for_user.invalidate(
                self, cred['user_id'], None
            )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from keystone.common import resource_options
from keystone.tests import unit


class TestResourceOptionObjects(unit.BaseTestCase):
    def test_option_init_validation(self):
        # option_name must be a string
        self.assertRaises(
            TypeError, resource_options.ResourceOption, 'test', 1234
        )
        # option_id must be a string
        self.assertRaises(
            TypeError, resource_options.ResourceOption, 1234, 'testing'
        )
        # option_id must be 4 characters
        self.assertRaises(
            ValueError, resource_options.ResourceOption, 'testing', 'testing'
        )
        resource_options.ResourceOption('test', 'testing')

    def test_duplicate_option_cases(self):
        option_id_str_valid = 'test'
        registry = resource_options.ResourceOptionRegistry(option_id_str_valid)
        option_name_unique = uuid.uuid4().hex

        option = resource_options.ResourceOption(
            option_id_str_valid, option_name_unique
        )
        option_dup_id = resource_options.ResourceOption(
            option_id_str_valid, uuid.uuid4().hex
        )
        option_dup_name = resource_options.ResourceOption(
            uuid.uuid4().hex[:4], option_name_unique
        )

        registry.register_option(option)

        self.assertRaises(ValueError, registry.register_option, option_dup_id)
        self.assertRaises(
            ValueError, registry.register_option, option_dup_name
        )
        self.assertIs(1, len(registry.options))
        registry.register_option(option)
        self.assertIs(1, len(registry.options))

    def test_registry(self):
        option = resource_options.ResourceOption(
            uuid.uuid4().hex[:4], uuid.uuid4().hex
        )
        option2 = resource_options.ResourceOption(
            uuid.uuid4().hex[:4], uuid.uuid4().hex
        )
        registry = resource_options.ResourceOptionRegistry('TEST')

        registry.register_option(option)
        self.assertIn(option.option_name, registry.option_names)
        self.assertIs(1, len(registry.options))
        self.assertIn(option.option_id, registry.option_ids)
        registry.register_option(option2)
        self.assertIn(option2.option_name, registry.option_names)
        self.assertIs(2, len(registry.options))
        self.assertIn(option2.option_id, registry.option_ids)
        self.assertIs(option, registry.get_option_by_id(option.option_id))
        self.assertIs(option2, registry.get_option_by_id(option2.option_id))
        self.assertIs(option, registry.get_option_by_name(option.option_name))
        self.assertIs(
            option2, registry.get_option_by_name(option2.option_name)
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import versionutils
from oslo_policy import policy

from keystone.common.policies import base

DEPRECATED_REASON = (
    "The EC2 credential API is now aware of system scope and default roles."
)

deprecated_ec2_get_credential = policy.DeprecatedRule(
    name=base.IDENTITY % 'ec2_get_credential',
    check_str=base.RULE_ADMIN_OR_CREDENTIAL_OWNER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)
deprecated_ec2_list_credentials = policy.DeprecatedRule(
    name=base.IDENTITY % 'ec2_list_credentials',
    check_str=base.RULE_ADMIN_OR_OWNER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)
deprecated_ec2_create_credential = policy.DeprecatedRule(
    name=base.IDENTITY % 'ec2_create_credential',
    check_str=base.RULE_ADMIN_OR_OWNER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)
deprecated_ec2_delete_credential = policy.DeprecatedRule(
    name=base.IDENTITY % 'ec2_delete_credential',
    check_str=base.RULE_ADMIN_OR_CREDENTIAL_OWNER,
    deprecated_reason=DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.TRAIN,
)


ec2_credential_policies = [
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'ec2_get_credential',
        check_str=base.ADMIN_OR_SYSTEM_READER_OR_CRED_OWNER,
        scope_types=['system', 'project'],
        description='Show ec2 credential details.',
        operations=[
            {
                'path': (
                    '/v3/users/{user_id}/credentials/OS-EC2/{credential_id}'
                ),
                'method': 'GET',
            }
        ],
        deprecated_rule=deprecated_ec2_get_credential,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'ec2_list_credentials',
        check_str=base.ADMIN_OR_SYSTEM_READER_OR_OWNER,
        scope_types=['system', 'project'],
        description='List ec2 credentials.',
        operations=[
            {'path': '/v3/users/{user_id}/credentials/OS-EC2', 'method': 'GET'}
        ],
        deprecated_rule=deprecated_ec2_list_credentials,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'ec2_create_credential',
        check_str=base.RULE_ADMIN_OR_OWNER,
        scope_types=['system', 'project'],
        description='Create ec2 credential.',
        operations=[
            {
                'path': '/v3/users/{user_id}/credentials/OS-EC2',
                'method': 'POST',
            }
        ],
        deprecated_rule=deprecated_ec2_create_credential,
    ),
    policy.DocumentedRuleDefault(
        name=base.IDENTITY % 'ec2_delete_credential',
        check_str=base.ADMIN_OR_CRED_OWNER,
        scope_types=['system', 'project'],
        description='Delete ec2 credential.',
        operations=[
            {
                'path': (
                    '/v3/users/{user_id}/credentials/OS-EC2/{credential_id}'
                ),
                'method': 'DELETE',
            }
        ],
        deprecated_rule=deprecated_ec2_delete_credential,
    ),
]


def list_rules():
    return ec2_credential_policies



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import http.client
import uuid

from oslo_serialization import jsonutils

from keystone.common.policies import project as pp
from keystone.common import provider_api
import keystone.conf
from keystone.tests.common import auth as common_auth
from keystone.tests import unit
from keystone.tests.unit import base_classes
from keystone.tests.unit import ksfixtures
from keystone.tests.unit.ksfixtures import temporaryfile

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


def _override_policy(policy_file):
    # TODO(lbragstad): Remove this once the deprecated policies in
    # keystone.common.policies.project have been removed. This is only
    # here to make sure we test the new policies instead of the deprecated
    # ones. Oslo.policy will OR deprecated policies with new policies to
    # maintain compatibility and give operators a chance to update
    # permissions or update policies without breaking users. This will
    # cause these specific tests to fail since we're trying to correct this
    # broken behavior with better scope checking.
    with open(policy_file, 'w') as f:
        overridden_policies = {
            'identity:get_project_tag': (
                pp.SYSTEM_READER_OR_DOMAIN_READER_OR_PROJECT_USER
            ),
            'identity:list_project_tags': (
                pp.SYSTEM_READER_OR_DOMAIN_READER_OR_PROJECT_USER
            ),
            'identity:create_project_tag': (
                pp.SYSTEM_ADMIN_OR_DOMAIN_ADMIN_OR_PROJECT_ADMIN
            ),
            'identity:update_project_tags': (
                pp.SYSTEM_ADMIN_OR_DOMAIN_ADMIN_OR_PROJECT_ADMIN
            ),
            'identity:delete_project_tag': (
                pp.SYSTEM_ADMIN_OR_DOMAIN_ADMIN_OR_PROJECT_ADMIN
            ),
            'identity:delete_project_tags': (
                pp.SYSTEM_ADMIN_OR_DOMAIN_ADMIN_OR_PROJECT_ADMIN
            ),
        }
        f.write(jsonutils.dumps(overridden_policies))


class _SystemUserTests:
    def test_user_can_get_project_tag(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_user_can_list_project_tags(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            r = c.get(
                '/v3/projects/{}/tags'.format(project['id']),
                headers=self.headers,
            )
            self.assertTrue(len(r.json['tags']) == 1)
            self.assertEqual(tag, r.json['tags'][0])


class _SystemMemberAndReaderTagTests:
    def test_user_cannot_create_project_tag(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_project_tag(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        update = {"tags": [uuid.uuid4().hex]}

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags'.format(project['id']),
                headers=self.headers,
                json=update,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_project_tag(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class _DomainAndProjectUserTagTests:
    def test_user_cannot_create_project_tag(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_project_tag(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        update = {"tags": [uuid.uuid4().hex]}

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags'.format(project['id']),
                headers=self.headers,
                json=update,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_project_tag(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class SystemReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserTests,
    _SystemMemberAndReaderTagTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_reader = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_reader)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.reader_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_reader['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserTests,
    _SystemMemberAndReaderTagTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        system_member = unit.new_user_ref(
            domain_id=CONF.identity.default_domain_id
        )
        self.user_id = PROVIDERS.identity_api.create_user(system_member)['id']
        PROVIDERS.assignment_api.create_system_grant_for_user(
            self.user_id, self.bootstrapper.member_role_id
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=system_member['password'],
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class SystemAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _SystemUserTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()
        self.useFixture(ksfixtures.Policy(self.config_fixture))
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        self.user_id = self.bootstrapper.admin_user_id
        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            system=True,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_create_project_tag(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.CREATED,
            )

    def test_user_can_update_project_tag(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        update = {"tags": [uuid.uuid4().hex]}

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags'.format(project['id']),
                headers=self.headers,
                json=update,
                expected_status_code=http.client.OK,
            )

    def test_user_can_delete_project_tag(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
            )


class _DomainUserTagTests:
    def test_user_can_get_tag_for_project_in_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_user_can_list_tags_for_project_in_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            r = c.get(
                '/v3/projects/{}/tags'.format(project['id']),
                headers=self.headers,
            )
            self.assertTrue(len(r.json['tags']) == 1)
            self.assertEqual(tag, r.json['tags'][0])

    def test_user_cannot_create_project_tag_outside_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_project_tag_outside_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        new_tag = uuid.uuid4().hex
        update = {"tags": [new_tag]}

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags'.format(project['id']),
                headers=self.headers,
                json=update,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_project_tag_outside_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_get_tag_for_project_outside_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_list_tags_for_project_outside_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/tags'.format(project['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class _DomainMemberAndReaderTagTests:
    def test_user_cannot_create_project_tag_in_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        tag = uuid.uuid4().hex

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_project_tag_in_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        new_tag = uuid.uuid4().hex
        update = {"tags": [new_tag]}

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags'.format(project['id']),
                headers=self.headers,
                json=update,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_project_tag_in_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class DomainAdminUserTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainUserTagTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        _override_policy(self.policy_file_name)
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_admin = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.admin_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_create_project_tag_in_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        tag = uuid.uuid4().hex

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.CREATED,
            )

    def test_user_can_update_project_tag_in_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        new_tag = uuid.uuid4().hex
        update = {"tags": [new_tag]}

        with self.test_client() as c:
            r = c.put(
                '/v3/projects/{}/tags'.format(project['id']),
                headers=self.headers,
                json=update,
                expected_status_code=http.client.OK,
            )
            self.assertTrue(len(r.json['tags']) == 1)
            self.assertEqual(new_tag, r.json['tags'][0])

    def test_user_can_delete_project_tag_in_domain(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex, unit.new_project_ref(domain_id=self.domain_id)
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
            )


class DomainMemberUserTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainUserTagTests,
    _DomainMemberAndReaderTagTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        _override_policy(self.policy_file_name)
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_admin = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class DomainReaderUserTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _DomainUserTagTests,
    _DomainMemberAndReaderTagTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )

        _override_policy(self.policy_file_name)
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        domain = PROVIDERS.resource_api.create_domain(
            uuid.uuid4().hex, unit.new_domain_ref()
        )
        self.domain_id = domain['id']
        domain_admin = unit.new_user_ref(domain_id=self.domain_id)
        self.user_id = PROVIDERS.identity_api.create_user(domain_admin)['id']
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=self.user_id,
            domain_id=self.domain_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=domain_admin['password'],
            domain_id=self.domain_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class _ProjectUserTagTests:
    def test_user_can_get_tag_for_project(self):
        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(self.project_id, tag)

        with self.test_client() as c:
            c.get(
                f'/v3/projects/{self.project_id}/tags/{tag}',
                headers=self.headers,
                expected_status_code=http.client.NO_CONTENT,
            )

    def test_user_can_list_tags_for_project(self):
        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(self.project_id, tag)

        with self.test_client() as c:
            r = c.get(
                f'/v3/projects/{self.project_id}/tags', headers=self.headers
            )
            self.assertTrue(len(r.json['tags']) == 1)
            self.assertEqual(tag, r.json['tags'][0])

    def test_user_cannot_create_tag_for_other_project(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_tag_for_other_project(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        update = {"tags": [uuid.uuid4().hex]}

        with self.test_client() as c:
            c.put(
                '/v3/projects/{}/tags'.format(project['id']),
                headers=self.headers,
                json=update,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_tag_for_other_project(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.delete(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_get_tag_for_other_project(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/tags/{}'.format(project['id'], tag),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_list_tags_for_other_project(self):
        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )

        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(project['id'], tag)

        with self.test_client() as c:
            c.get(
                '/v3/projects/{}/tags'.format(project['id']),
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class _ProjectMemberAndReaderTagTests:
    def test_user_cannot_create_project_tag(self):
        tag = uuid.uuid4().hex
        with self.test_client() as c:
            c.put(
                f'/v3/projects/{self.project_id}/tags/{tag}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_update_project_tag(self):
        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(self.project_id, tag)

        update = {"tags": [uuid.uuid4().hex]}

        with self.test_client() as c:
            c.put(
                f'/v3/projects/{self.project_id}/tags',
                headers=self.headers,
                json=update,
                expected_status_code=http.client.FORBIDDEN,
            )

    def test_user_cannot_delete_project_tag(self):
        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(self.project_id, tag)

        with self.test_client() as c:
            c.delete(
                f'/v3/projects/{self.project_id}/tags/{tag}',
                headers=self.headers,
                expected_status_code=http.client.FORBIDDEN,
            )


class ProjectAdminTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _ProjectUserTagTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        _override_policy(self.policy_file_name)
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        self.user_id = self.bootstrapper.admin_user_id
        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.admin_role_id,
            user_id=self.user_id,
            project_id=self.bootstrapper.project_id,
        )
        self.project_id = self.bootstrapper.project_id

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            project_id=self.bootstrapper.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}

    def test_user_can_create_project_tag(self):
        tag = uuid.uuid4().hex
        with self.test_client() as c:
            c.put(
                f'/v3/projects/{self.project_id}/tags/{tag}',
                headers=self.headers,
                expected_status_code=http.client.CREATED,
            )

    def test_user_can_update_project_tag(self):
        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(self.project_id, tag)

        update = {"tags": [uuid.uuid4().hex]}

        with self.test_client() as c:
            c.put(
                f'/v3/projects/{self.project_id}/tags',
                headers=self.headers,
                json=update,
                expected_status_code=http.client.OK,
            )

    def test_user_can_delete_project_tag(self):
        tag = uuid.uuid4().hex
        PROVIDERS.resource_api.create_project_tag(self.project_id, tag)

        with self.test_client() as c:
            c.delete(
                f'/v3/projects/{self.project_id}/tags/{tag}',
                headers=self.headers,
            )


class ProjectMemberTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _ProjectUserTagTests,
    _ProjectMemberAndReaderTagTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        _override_policy(self.policy_file_name)
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )
        self.project_id = project['id']
        self.user_id = self.bootstrapper.admin_user_id

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.member_role_id,
            user_id=self.user_id,
            project_id=self.project_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            project_id=self.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}


class ProjectReaderTests(
    base_classes.TestCaseWithBootstrap,
    common_auth.AuthTestMixin,
    _ProjectUserTagTests,
    _ProjectMemberAndReaderTagTests,
):
    def setUp(self):
        super().setUp()
        self.loadapp()

        self.policy_file = self.useFixture(temporaryfile.SecureTempFile())
        self.policy_file_name = self.policy_file.file_name
        self.useFixture(
            ksfixtures.Policy(
                self.config_fixture, policy_file=self.policy_file_name
            )
        )
        _override_policy(self.policy_file_name)
        self.config_fixture.config(group='oslo_policy', enforce_scope=True)

        project = PROVIDERS.resource_api.create_project(
            uuid.uuid4().hex,
            unit.new_project_ref(domain_id=CONF.identity.default_domain_id),
        )
        self.project_id = project['id']
        self.user_id = self.bootstrapper.admin_user_id

        PROVIDERS.assignment_api.create_grant(
            self.bootstrapper.reader_role_id,
            user_id=self.user_id,
            project_id=self.project_id,
        )

        auth = self.build_authentication_request(
            user_id=self.user_id,
            password=self.bootstrapper.admin_password,
            project_id=self.project_id,
        )

        # Grab a token using the persona we're testing and prepare headers
        # for requests we'll be making in the tests.
        with self.test_client() as c:
            r = c.post('/v3/auth/tokens', json=auth)
            self.token_id = r.headers['X-Subject-Token']
            self.headers = {'X-Auth-Token': self.token_id}



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import uuid

from keystone.common import jwt_utils
from keystone.common import provider_api
from keystone.common import utils
import keystone.conf
from keystone import exception
from keystone.models import token_model
from keystone.tests import unit
from keystone.tests.unit import ksfixtures
from keystone.token import provider
from keystone.token.providers import jws

CONF = keystone.conf.CONF
PROVIDERS = provider_api.ProviderAPIs


class TestJWSProvider(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.config_fixture.config(group='token', provider='jws')
        self.useFixture(ksfixtures.JWSKeyRepository(self.config_fixture))
        self.provider = jws.Provider()

    def test_invalid_token_raises_token_not_found(self):
        token_id = uuid.uuid4().hex
        self.assertRaises(
            exception.TokenNotFound, self.provider.validate_token, token_id
        )

    def test_non_existent_private_key_raises_system_exception(self):
        private_key = os.path.join(
            CONF.jwt_tokens.jws_private_key_repository, 'private.pem'
        )
        os.remove(private_key)
        self.assertRaises(SystemExit, jws.Provider)

    def test_non_existent_public_key_repo_raises_system_exception(self):
        for f in os.listdir(CONF.jwt_tokens.jws_public_key_repository):
            path = os.path.join(CONF.jwt_tokens.jws_public_key_repository, f)
            os.remove(path)
        os.rmdir(CONF.jwt_tokens.jws_public_key_repository)
        self.assertRaises(SystemExit, jws.Provider)

    def test_empty_public_key_repo_raises_system_exception(self):
        for f in os.listdir(CONF.jwt_tokens.jws_public_key_repository):
            path = os.path.join(CONF.jwt_tokens.jws_public_key_repository, f)
            os.remove(path)
        self.assertRaises(SystemExit, jws.Provider)

    def test_unable_to_verify_token_with_missing_public_key(self):
        # create token, signing with private key
        token = token_model.TokenModel()
        token.methods = ['password']
        token.user_id = uuid.uuid4().hex
        token.audit_id = provider.random_urlsafe_str()
        token.expires_at = utils.isotime(
            provider.default_expire_time(), subsecond=True
        )
        token_id, issued_at = self.provider.generate_id_and_issued_at(token)

        # remove the public key for the token we just created
        current_pub_key = os.path.join(
            CONF.jwt_tokens.jws_public_key_repository, 'public.pem'
        )
        os.remove(current_pub_key)

        # create additional public keys
        for _ in range(2):
            private_key_path = os.path.join(
                CONF.jwt_tokens.jws_private_key_repository, uuid.uuid4().hex
            )
            pub_key_path = os.path.join(
                CONF.jwt_tokens.jws_public_key_repository, uuid.uuid4().hex
            )
            jwt_utils.create_jws_keypair(private_key_path, pub_key_path)

        # validate token and ensure it returns a 404
        self.assertRaises(
            exception.TokenNotFound, self.provider.validate_token, token_id
        )

    def test_verify_token_with_multiple_public_keys_present(self):
        token = token_model.TokenModel()
        token.methods = ['password']
        token.user_id = uuid.uuid4().hex
        token.audit_id = provider.random_urlsafe_str()
        token.expires_at = utils.isotime(
            provider.default_expire_time(), subsecond=True
        )
        token_id, issued_at = self.provider.generate_id_and_issued_at(token)

        for _ in range(2):
            private_key_path = os.path.join(
                CONF.jwt_tokens.jws_private_key_repository, uuid.uuid4().hex
            )
            pub_key_path = os.path.join(
                CONF.jwt_tokens.jws_public_key_repository, uuid.uuid4().hex
            )
            jwt_utils.create_jws_keypair(private_key_path, pub_key_path)

        # make sure we iterate through all public keys on disk and we can still
        # validate the token
        self.provider.validate_token(token_id)



# Copyright 2012 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
Test for SQL migration extensions.

To run these tests against a live database:

1. Set up a blank, live database.
2. Export database information to environment variable
   ``OS_TEST_DBAPI_ADMIN_CONNECTION``. For example::

    export OS_TEST_DBAPI_ADMIN_CONNECTION=postgresql://localhost/postgres?host=
    /var/folders/7k/pwdhb_mj2cv4zyr0kyrlzjx40000gq/T/tmpMGqN8C&port=9824

3. Run the tests using::

    tox -e py39 -- keystone.tests.unit.test_sql_upgrade

For further information, see `oslo.db documentation
<https://docs.openstack.org/oslo.db/latest/contributor/index.html#how-to-run-unit-tests>`_.

.. warning::

    Your database will be wiped.

    Do not do this against a database with valuable data as
    all data will be lost.
"""

import fixtures
from oslo_db import options as db_options
from oslo_db.sqlalchemy import enginefacade
from oslo_db.sqlalchemy import test_fixtures as db_fixtures
from oslo_log import fixture as log_fixture
from oslo_log import log
import sqlalchemy.exc

from keystone.cmd import cli
from keystone.common import sql
from keystone.common.sql import upgrades
import keystone.conf
from keystone.tests import unit
from keystone.tests.unit import ksfixtures

CONF = keystone.conf.CONF

# NOTE(morganfainberg): This should be updated when each DB migration collapse
# is done to mirror the expected structure of the DB in the format of
# { <DB_TABLE_NAME>: [<COLUMN>, <COLUMN>, ...], ... }
INITIAL_TABLE_STRUCTURE = {
    'config_register': ['type', 'domain_id'],
    'credential': [
        'id',
        'user_id',
        'project_id',
        'type',
        'extra',
        'key_hash',
        'encrypted_blob',
    ],
    'endpoint': [
        'id',
        'legacy_endpoint_id',
        'interface',
        'region_id',
        'service_id',
        'url',
        'enabled',
        'extra',
    ],
    'group': ['id', 'domain_id', 'name', 'description', 'extra'],
    'policy': ['id', 'type', 'blob', 'extra'],
    'project': [
        'id',
        'name',
        'extra',
        'description',
        'enabled',
        'domain_id',
        'parent_id',
        'is_domain',
    ],
    'project_option': ['project_id', 'option_id', 'option_value'],
    'project_tag': ['project_id', 'name'],
    'role': ['id', 'name', 'extra', 'domain_id', 'description'],
    'role_option': ['role_id', 'option_id', 'option_value'],
    'service': ['id', 'type', 'extra', 'enabled'],
    'token': ['id', 'expires', 'extra', 'valid', 'trust_id', 'user_id'],
    'trust': [
        'id',
        'trustor_user_id',
        'trustee_user_id',
        'project_id',
        'impersonation',
        'deleted_at',
        'expires_at',
        'remaining_uses',
        'extra',
        'expires_at_int',
        'redelegated_trust_id',
        'redelegation_count',
    ],
    'trust_role': ['trust_id', 'role_id'],
    'user': [
        'id',
        'extra',
        'enabled',
        'default_project_id',
        'created_at',
        'last_active_at',
        'domain_id',
    ],
    'user_option': ['user_id', 'option_id', 'option_value'],
    'user_group_membership': ['user_id', 'group_id'],
    'region': ['id', 'description', 'parent_region_id', 'extra'],
    'assignment': ['type', 'actor_id', 'target_id', 'role_id', 'inherited'],
    'id_mapping': ['public_id', 'domain_id', 'local_id', 'entity_type'],
    'whitelisted_config': ['domain_id', 'group', 'option', 'value'],
    'sensitive_config': ['domain_id', 'group', 'option', 'value'],
    'policy_association': [
        'id',
        'policy_id',
        'endpoint_id',
        'service_id',
        'region_id',
    ],
    'identity_provider': [
        'id',
        'enabled',
        'description',
        'domain_id',
        'authorization_ttl',
    ],
    'federation_protocol': [
        'id',
        'idp_id',
        'mapping_id',
        'remote_id_attribute',
    ],
    'mapping': ['id', 'rules', 'schema_version'],
    'service_provider': [
        'auth_url',
        'id',
        'enabled',
        'description',
        'sp_url',
        'relay_state_prefix',
    ],
    'idp_remote_ids': ['idp_id', 'remote_id'],
    'consumer': ['id', 'description', 'secret', 'extra'],
    'request_token': [
        'id',
        'request_secret',
        'verifier',
        'authorizing_user_id',
        'requested_project_id',
        'role_ids',
        'consumer_id',
        'expires_at',
    ],
    'access_token': [
        'id',
        'access_secret',
        'authorizing_user_id',
        'project_id',
        'role_ids',
        'consumer_id',
        'expires_at',
    ],
    'revocation_event': [
        'id',
        'domain_id',
        'project_id',
        'user_id',
        'role_id',
        'trust_id',
        'consumer_id',
        'access_token_id',
        'issued_before',
        'expires_at',
        'revoked_at',
        'audit_id',
        'audit_chain_id',
    ],
    'project_endpoint': ['endpoint_id', 'project_id'],
    'endpoint_group': ['id', 'name', 'description', 'filters'],
    'project_endpoint_group': ['endpoint_group_id', 'project_id'],
    'implied_role': ['prior_role_id', 'implied_role_id'],
    'local_user': [
        'id',
        'user_id',
        'domain_id',
        'name',
        'failed_auth_count',
        'failed_auth_at',
    ],
    'password': [
        'id',
        'local_user_id',
        'created_at',
        'expires_at',
        'self_service',
        'password_hash',
        'created_at_int',
        'expires_at_int',
    ],
    'federated_user': [
        'id',
        'user_id',
        'idp_id',
        'protocol_id',
        'unique_id',
        'display_name',
    ],
    'nonlocal_user': ['domain_id', 'name', 'user_id'],
    'system_assignment': [
        'type',
        'actor_id',
        'target_id',
        'role_id',
        'inherited',
    ],
    'registered_limit': [
        'internal_id',
        'id',
        'service_id',
        'region_id',
        'resource_name',
        'default_limit',
        'description',
    ],
    'limit': [
        'internal_id',
        'id',
        'project_id',
        'resource_limit',
        'description',
        'registered_limit_id',
        'domain_id',
    ],
    'application_credential': [
        'internal_id',
        'id',
        'name',
        'secret_hash',
        'description',
        'user_id',
        'project_id',
        'expires_at',
        'system',
        'unrestricted',
    ],
    'application_credential_role': ['application_credential_id', 'role_id'],
    'access_rule': [
        'id',
        'service',
        'path',
        'method',
        'external_id',
        'user_id',
    ],
    'application_credential_access_rule': [
        'application_credential_id',
        'access_rule_id',
    ],
    'expiring_user_group_membership': [
        'user_id',
        'group_id',
        'idp_id',
        'last_verified',
    ],
}


class MigrateBase(db_fixtures.OpportunisticDBTestMixin):
    """Test complete orchestration between all database phases."""

    def setUp(self):
        super().setUp()

        self.useFixture(log_fixture.get_logging_handle_error_fixture())
        self.stdlog = self.useFixture(ksfixtures.StandardLogging())
        self.useFixture(ksfixtures.WarningsFixture())

        self.engine = enginefacade.writer.get_engine()
        self.sessionmaker = enginefacade.writer.get_sessionmaker()

        db_options.set_defaults(CONF, connection=self.engine.url)

        # Override keystone's context manager to be oslo.db's global context
        # manager.
        sql.core._TESTING_USE_GLOBAL_CONTEXT_MANAGER = True
        self.addCleanup(
            setattr, sql.core, '_TESTING_USE_GLOBAL_CONTEXT_MANAGER', False
        )
        self.addCleanup(sql.cleanup)

    def expand(self):
        """Expand database schema."""
        upgrades.expand_schema(engine=self.engine)

    def contract(self):
        """Contract database schema."""
        upgrades.contract_schema(engine=self.engine)

    @property
    def metadata(self):
        """A collection of tables and their associated schemas."""
        return sqlalchemy.MetaData()

    def load_table(self, name):
        table = sqlalchemy.Table(
            name, self.metadata, autoload_with=self.engine
        )
        return table

    def assertTableDoesNotExist(self, table_name):
        """Assert that a given table exists cannot be selected by name."""
        # Switch to a different metadata otherwise you might still
        # detect renamed or dropped tables
        try:
            sqlalchemy.Table(
                table_name, self.metadata, autoload_with=self.engine
            )
        except sqlalchemy.exc.NoSuchTableError:
            pass
        else:
            raise AssertionError(f'Table "{table_name}" already exists')

    def assertTableColumns(self, table_name, expected_cols):
        """Assert that the table contains the expected set of columns."""
        table = self.load_table(table_name)
        actual_cols = [col.name for col in table.columns]
        # Check if the columns are equal, but allow for a different order,
        # which might occur after an upgrade followed by a downgrade
        self.assertCountEqual(
            expected_cols, actual_cols, f'{table_name} table'
        )

    def test_db_sync_check(self):
        checker = cli.DbSync()

        # If the expand repository doesn't exist yet, then we need to make sure
        # we advertise that `--expand` must be run first.
        log_info = self.useFixture(fixtures.FakeLogger(level=log.INFO))
        status = checker.check_db_sync_status()
        self.assertIn("keystone-manage db_sync --expand", log_info.output)
        self.assertEqual(status, 2)

        # Assert the correct message is printed when migrate is ahead of
        # contract
        self.expand()
        log_info = self.useFixture(fixtures.FakeLogger(level=log.INFO))
        status = checker.check_db_sync_status()
        self.assertIn("keystone-manage db_sync --contract", log_info.output)
        self.assertEqual(status, 4)

        # Assert the correct message gets printed when all commands are on
        # the same version
        self.contract()
        log_info = self.useFixture(fixtures.FakeLogger(level=log.INFO))
        status = checker.check_db_sync_status()
        self.assertIn("All db_sync commands are upgraded", log_info.output)
        self.assertEqual(status, 0)

    def test_upgrade_add_initial_tables(self):
        self.expand()
        for table in INITIAL_TABLE_STRUCTURE:
            self.assertTableColumns(table, INITIAL_TABLE_STRUCTURE[table])


class FullMigrationSQLite(MigrateBase, unit.TestCase):
    pass


class FullMigrationMySQL(MigrateBase, unit.TestCase):
    FIXTURE = db_fixtures.MySQLOpportunisticFixture


class FullMigrationPostgreSQL(MigrateBase, unit.TestCase):
    FIXTURE = db_fixtures.PostgresqlOpportunisticFixture
